_.italic  # unused attribute (composer/docx.py:115)
_.small_caps  # unused attribute (composer/docx.py:117)
_.style  # unused attribute (composer/docx.py:119)
_.style  # unused attribute (composer/docx.py:125)
_.space_before  # unused attribute (composer/docx.py:139)
_.keep_together  # unused attribute (composer/docx.py:140)
_.bold  # unused attribute (composer/docx.py:145)
_.keep_together  # unused attribute (composer/docx.py:186)
_.space_before  # unused attribute (composer/docx.py:187)
_.bold  # unused attribute (composer/docx.py:188)
_.keep_together  # unused attribute (composer/docx.py:195)
_.bold  # unused attribute (composer/docx.py:198)
_.space_before  # unused attribute (composer/docx.py:222)
_.page_break_before  # unused attribute (composer/docx.py:235)
_.page_break_before  # unused attribute (composer/docx.py:238)
_.alignment  # unused attribute (composer/docx.py:241)
_.keep_with_next  # unused attribute (composer/docx.py:242)
_.bold  # unused attribute (composer/docx.py:249)
_.bold  # unused attribute (composer/docx.py:256)
_.italic  # unused attribute (composer/pptx.py:90)
_.word_wrap  # unused attribute (composer/pptx.py:134)
_.rgb  # unused attribute (composer/pptx.py:202)
_.word_wrap  # unused attribute (composer/pptx.py:329)
_.word_wrap  # unused attribute (composer/pptx.py:345)
_.word_wrap  # unused attribute (composer/pptx.py:402)
_.bold  # unused attribute (composer/pptx.py:418)
_.bold  # unused attribute (composer/pptx.py:426)
_.bold  # unused attribute (composer/pptx.py:432)
_.body_width  # unused attribute (parser.py:800)
tokens  # unused variable (parser_db.py:18)
states  # unused variable (parser_db.py:37)
t_TITLE  # unused function (parser_db.py:81)
t_URL  # unused function (parser_db.py:87)
t_DATE  # unused function (parser_db.py:93)
t_EDITOR  # unused function (parser_db.py:99)
t_INFO  # unused function (parser_db.py:105)
t_TOUR  # unused function (parser_db.py:111)
t_QUESTION  # unused function (parser_db.py:118)
t_ANSWER  # unused function (parser_db.py:125)
t_ZACHET  # unused function (parser_db.py:136)
t_NEZACHET  # unused function (parser_db.py:142)
t_COMMENT  # unused function (parser_db.py:148)
t_SOURCE  # unused function (parser_db.py:159)
t_AUTHOR  # unused function (parser_db.py:170)
t_title_end  # unused function (parser_db.py:176)
t_url_end  # unused function (parser_db.py:183)
t_date_end  # unused function (parser_db.py:189)
t_info_end  # unused function (parser_db.py:195)
t_editor_end  # unused function (parser_db.py:201)
t_tour_end  # unused function (parser_db.py:207)
t_question_HANDOUT  # unused function (parser_db.py:213)
t_handout_end  # unused function (parser_db.py:219)
t_question_PIC  # unused function (parser_db.py:225)
t_question_TEXT  # unused function (parser_db.py:230)
t_question_end  # unused function (parser_db.py:251)
t_answer_TEXT  # unused function (parser_db.py:261)
t_answer_end  # unused function (parser_db.py:275)
t_zachet_end  # unused function (parser_db.py:289)
t_nezachet_end  # unused function (parser_db.py:295)
t_comment_TEXT  # unused function (parser_db.py:301)
t_comment_end  # unused function (parser_db.py:327)
t_source_TEXT  # unused function (parser_db.py:342)
t_source_end  # unused function (parser_db.py:360)
t_author_end  # unused function (parser_db.py:375)
t_ANY_TEXT  # unused function (parser_db.py:381)
t_ANY_ENDLINE  # unused function (parser_db.py:391)
t_ANY_error  # unused function (parser_db.py:396)
_.bold  # unused attribute (trello.py:356)
_.bold  # unused attribute (trello.py:362)
СYRILLIC_CHARS  # unused variable (typotools.py:58)
